/*
  UDP Session 
  by Joel R. Voss
  Jan 20, 2006
  Original Concept by Winston Williams, others
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <err.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

#include <errno.h>
extern int errno;

#include <time.h>

#include <ssl_bf2a.h>
#include <keyex1.h>

#ifdef _DEBUG
#define VERBOSE 1
#else
#define VERBOSE 0
#endif

typedef enum UDP_CONTROL
{
    UDP_UNKNOWN = 0,
    UDP_QUIT = 1,
    UDP_SEND = 2,
    UDP_SENDTO = 3,
    UDP_RECV = 4,
    UDP_RECVFROM = 5,
    UDP_CONNECT = 6,
    UDP_ERROR = 7
} UDP_CONTROL;

typedef enum UDP_AUTH
{
    UDP_AUTH_UNKNOWN = 0,
    UDP_AUTH_NULL = 1,
    UDP_AUTH_PASSWORD = 2,
    UDP_AUTH_DSA = 3,
} UDP_AUTH;

const char *udp_control_str(UDP_CONTROL c)
{
    switch(c)
    {
    case UDP_UNKNOWN:
        return "UDP Unknown";
        break;
    case UDP_QUIT:
        return "UDP Quit";
        break;
    case UDP_SEND:
        return "UDP Send";
        break;
    case UDP_SENDTO:
        return "UDP Send To";
        break;
    case UDP_RECV:
        return "UDP Receive";
        break;
    case UDP_RECVFROM:
        return "UDP Receive From";
        break;
    default:
        break;
    }
    return "UDP Strange";
}

UDP_CONTROL send_data(char *data, int *data_size, int s, struct sockaddr *raddr)
{
    UDP_CONTROL c = UDP_SENDTO;
    sendto(s, data, *data_size, 0, raddr, sizeof(struct sockaddr_in));
    return c;
}

UDP_CONTROL recv_data(char *out, int *recv_size, int s, struct sockaddr *oaddr)
{
    UDP_CONTROL c = UDP_RECVFROM;
    socklen_t addr_size = sizeof(struct sockaddr_in);
    *recv_size = recvfrom(s, out, 1024+52, 0, oaddr, &addr_size);
	
	// Disregard any pokes. They could be sent at any time.
	if((*recv_size == 5) && (memcmp(out, "/poke", 5) == 0))
	{
		if(VERBOSE) fprintf(stderr, "Received a /poke, re-receiving.\n");
		// Do another recv.
		return recv_data(out, recv_size, s, oaddr);
	}
    return c;
}

/*
UDP_CONTROL send_file(FILE *data_in, int *data_size, int s, struct sockaddr *raddr)
{
    char data[1025];
    *data_size = fread(data, 1, 1024, data_in);
    //data[*data_size] = 0;
    return send_data(data, data_size, s, raddr);
}

UDP_CONTROL recv_file(FILE *data_out, int *recv_size, int s, struct sockaddr *oaddr)
{
    UDP_CONTROL c = UDP_RECVFROM;
    char out[1025];
    // The exit command is /quit.
    c = recv_data(out, recv_size, s, oaddr);
    //out[*recv_size] = 0;
    if(c == UDP_RECVFROM)
    {
        fwrite(out, *recv_size, 1, data_out);
    }
    return c;
}
*/

UDP_CONTROL send_poke(int s, struct sockaddr_in *raddr)
{
	// Poke a hole in your firewall first. Not encrypted, maybe it should be.
    char poke_data[] = "/poke";
    int data_size = strlen(poke_data);
    UDP_CONTROL c = send_data(poke_data, &data_size, s, (struct sockaddr *)raddr);
    if(VERBOSE) fprintf(stderr, "Poke %s (%i)\n", udp_control_str(c), data_size);
	return c;
}

BIGNUM *recv_dh_public_key(int s, struct sockaddr_in *raddr)
{
	BYTE pub_buf[1024+52];
	int recv_size = -1;
	//UDP_CONTROL c = 
	UDP_CONTROL c = recv_data(pub_buf, &recv_size, s, (struct sockaddr *)raddr);
	
	// TODO: verify header. re-receive.
	if(recv_size <= 0)
	{
		perror("recv dh");
		fprintf(stderr, "received dh %s (%i)\n", udp_control_str(c), recv_size);
		return 0;
	}
	
	if(memcmp(pub_buf, "dhkeyex_", 8) != 0)
	{
		fprintf(stderr, "bad dh header %s (%i)\n", udp_control_str(c), recv_size);
		return 0;
	}
	
	BIGNUM *pub_key;
	/* Converting from binary to BIGNUM */
	pub_key = BN_bin2bn(pub_buf + 8, recv_size - 8, NULL);
    
    return pub_key;
}

void send_public_key(BIGNUM *pub_key, int s, struct sockaddr_in *raddr)
{
	/* Converting from BIGNUM to binary */
	int len = BN_num_bytes(pub_key);
	BYTE *buf = (BYTE *)malloc(len + 8);
	memcpy(buf, "dhkeyex_", 8);
	len = BN_bn2bin(pub_key, buf + 8) + 8;
	
	// TODO: add header, re-send.
	//UDP_CONTROL c = 
	send_data(buf, &len, s, (struct sockaddr *)raddr);
	//return c;
	free(buf);
}

int dh_keyex(BYTE *secret, int *secret_len, int s, struct sockaddr_in *raddr)
{
    // Secret Key Exchange using Diffie-Hellman
	
	BIGNUM *dh_pub_key = 0;
    DH *dh;
    // Generate a key pair
    int ret = dh_gen_key_pair(&dh_pub_key, &dh);
	if(ret != 1) return 1;
    // Send the public key.
    send_public_key(dh_pub_key, s, raddr);
    // Receive the peer's public key.
    BIGNUM *other_dh_pub_key = recv_dh_public_key(s, raddr);
	if(other_dh_pub_key == 0) return 2;
    // Create the secret key with this data.
    ret = dh_gen_secret(dh, other_dh_pub_key, secret, secret_len);
    
    BN_free(other_dh_pub_key);
    BN_free(dh_pub_key);

    return ret;
}

int authenticate(const BYTE *secret, int secret_len, const char *password, int s, struct sockaddr_in *raddr, UDP_AUTH auth)
{
	if(auth == UDP_AUTH_PASSWORD)
	{
		// Password Auth.
		time_t utc_time_data = time(0);
		// time_t is a long, 32-bits.
		BYTE key[32];
		BYTE salt[8];
		// Throwaway variable.
		BYTE iv[EVP_MAX_IV_LENGTH];
		// First 4 bytes of salt are time. Next 4 bytes are random.
		memcpy(salt, &utc_time_data, 4);
		select_random_salt(salt + 4, 4);
		generate_key(key, password, salt, iv, KEY_LENGTH);
        char data[1024 + 52];
		int data_size = KEY_LENGTH + 16;
		memcpy(data, "passauth", 8);
		memcpy(data+8, salt, 8);
		memcpy(data+16, key, KEY_LENGTH);
		// Send the salt and the key.
        char enc_data[1024 + 52];
        int enc_size = -1;
        encrypt_packet(data, data_size, secret, secret_len, enc_data, &enc_size);
        UDP_CONTROL c = send_data(enc_data, &enc_size, s, (struct sockaddr *)raddr);
        if(VERBOSE) fprintf(stderr, "Sent %s (%i, %i)\n", udp_control_str(c), enc_size, data_size);

		// Check for password auth packet as the first packet.
        // Receive
        int recv_size = -1;
        c = recv_data(enc_data, &recv_size, s, (struct sockaddr *)raddr);
        if(VERBOSE) fprintf(stderr, "Recv %s (%i)\n", udp_control_str(c), recv_size);
        // TODO: Check for Salted__ before you decrypt?
        decrypt_packet(enc_data, recv_size, secret, secret_len, data, &data_size);
		// Ensure that the data is the correct size.
		if(data_size != KEY_LENGTH + 16)
		{
			fprintf(stderr, "Error: Recv password length %s (%i)\n", udp_control_str(c), data_size);
			return 1;
		}
		// Check the header.
		if(memcmp(data, "passauth", 8) != 0)
		{
			fprintf(stderr, "Error: Recv password header %8.8s\n", data);
			return 1;
		}
		BYTE key_recv[32];
		BYTE *salt_recv = data + 8;
		generate_key(key_recv, password, salt_recv, iv, KEY_LENGTH);
		if(memcmp(data + 16, key_recv, KEY_LENGTH) != 0)
		{
			fprintf(stderr, "Error: Recv password hash\n");
			return 1;
		}
		// Check the date to ensure that it is within acceptable values.
		time_t *recv_time_data = (time_t *)salt_recv;
		int diff = utc_time_data - *recv_time_data;
		if(diff < 0) diff *= -1;
		if(diff > 100)
		{
			fprintf(stderr, "Error: Recv password time is toofar off: %i secs.\n", 
				diff);
			return 1;
		}
		fprintf(stderr, "Password authentication succeeded.\n");
	}
	else if(auth == UDP_AUTH_DSA)
	{
		// DSA Auth.
		// TODO: Check for dsa auth packet as the first packet.
	}
	else
	{
		// Null Auth.
		// TODO: Check for null auth packet as the first packet.
	}
    return 0;
}

int chat_main(int s, struct sockaddr_in *saddr, struct sockaddr_in *raddr, UDP_AUTH auth)
{
    char password[256];
	BYTE secret[256];
	// Usually 256, but not required.
	int secret_len = 256;
    
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        return 1;
    }

	if(auth == UDP_AUTH_NULL)
	{
		fprintf(stdout, "Press any key: ");
		fread(password, 1, 1, stdin);
		fprintf(stdout, "\n");
	}
	else
	{
		// Secure password retreival algorithm from enc
		get_password(password, 256, 1);
	}

	send_poke(s, raddr);
    
	int ret = dh_keyex(secret, &secret_len, s, raddr);
	if(ret != 0)
	{
		// This is a rare error condition, likely due to DH_size being too big.
		return ret;
	}
	//print_hex(secret, secret_len);

	ret = authenticate(secret, secret_len, password, s, raddr, auth);
	if(ret != 0)
	{
		return ret;
	}
    
	int bEndChat = 0;

    // Chat
    while(1)
    {
        // Send
        int data_size = -1;
        char data[1025];
        fgets(data, 1024, stdin);
        // fgets always puts a \0 after the last char.
        // fgets never reads more than size.
        data_size = strlen(data);
        // The exit command is /quit.
        if(strncmp(data, "/quit", 6) == 0)
        {
            fprintf(stderr, "Got a specific quit command, sending to peer.\n");
            bEndChat = 1;
        }
        if(feof(stdin))
        {
            fprintf(stderr, "Got an EOF, sending /quit to peer.\n");
            strncpy(data, "/quit", 6);
			data_size = 5;
            bEndChat = 1;
        }
        char enc_data[1024 + 52];
        int enc_size = -1;
        encrypt_packet(data, data_size, secret, secret_len, enc_data, &enc_size);
        UDP_CONTROL c = send_data(enc_data, &enc_size, s, (struct sockaddr *)raddr);
        if(VERBOSE) fprintf(stderr, "Sent %s (%i, %i)\n", udp_control_str(c), enc_size, data_size);

        // If user initiated an end of chat (EOF or /quit), 
        // quit after sending it.        
        if(bEndChat)
        {
            break;
        }
        if(data_size <= 0 || c != UDP_SENDTO)
        {
            fprintf(stderr, "Error: Sent %s (%i)\n", udp_control_str(c), enc_size);
            perror("sendto");
            break;
        }
        // Receive
        int recv_size = -1;
        c = recv_data(enc_data, &recv_size, s, (struct sockaddr *)raddr);
        if(VERBOSE) fprintf(stderr, "Recv %s (%i)\n", udp_control_str(c), recv_size);
        if(recv_size <= 0 || c != UDP_RECVFROM)
        {
            fprintf(stderr, "Error: Received %s (%i)\n", udp_control_str(c), recv_size);
            perror("recvfrom");
            break;
        }
        // TODO: Check for Salted__ before you decrypt?
        decrypt_packet(enc_data, recv_size, secret, secret_len, data, &data_size);
        // Since we're printing it, it's important that it's zero ended. 
        // decrypt_packet _DOES NOT_ terminate the data with null.
        data[data_size] = 0;
        fwrite(data, data_size, 1, stdout);
        if((data_size == 5) && (strncmp(data, "/quit", 5) == 0))
        {
            if(VERBOSE) fprintf(stderr, "Got a specific remote quit command.\n");
			fprintf(stdout, "\n");
            break;
        }
    }
    return 0;
}

int send_main(FILE *data_io, int s, struct sockaddr_in *saddr, struct sockaddr_in *raddr, UDP_AUTH auth)
{
    char password[256];
	BYTE secret[256];
	// Usually 256, but not required.
	int secret_len = 256;

    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        return 1;
    }
    
	if(auth == UDP_AUTH_NULL)
	{
		fprintf(stdout, "Press any key: ");
		fread(password, 1, 1, stdin);
		fprintf(stdout, "\n");
	}
	else
	{
		// Secure password retreival algorithm from enc
		get_password(password, 256, 1);
	}

	send_poke(s, raddr);
	int ret = dh_keyex(secret, &secret_len, s, raddr);
	if(ret != 0)
	{
		return ret;
	}

	ret = authenticate(secret, secret_len, password, s, raddr, auth);
	if(ret != 0)
	{
		return ret;
	}

	// Send file.
    while(1)
    {
        int data_size = -1;
        fprintf(stderr, "Input(1024): \n");
        char data[1024];
        data_size = fread(data, 1, 1024, data_io);
        int enc_size = -1;
        char enc_data[1024 + 52];
        encrypt_packet(data, data_size, secret, secret_len, enc_data, &enc_size);
        UDP_CONTROL c = send_data(enc_data, &enc_size, s, (struct sockaddr *)raddr);
        fprintf(stderr, "Sending (%i)\n", enc_size);
        //sleep(1);
        // If the file has ended, send /quit.
        if(feof(data_io))
            break;
        if(data_size <= 0 || c != UDP_SENDTO)
        {
            fprintf(stderr, "Sent %s (%i)\n", udp_control_str(c), data_size);
            break;
        }
    }
    // End the file transfer with /quit.
    // Encrypt the quit of course.
    char data[6];
    int data_size = 5; // strlen(data);
    strncpy(data, "/quit", sizeof(data));
    int enc_size = -1;
    char enc_data[1024 + 52];
    encrypt_packet(data, data_size, secret, secret_len, enc_data, &enc_size);
    fprintf(stderr, "Sending %s (%i)\n", data, enc_size);
    send_data(enc_data, &enc_size, s, (struct sockaddr *)raddr);
    return 0;
}

int recv_main(FILE *data_io, int s, struct sockaddr_in *saddr, struct sockaddr_in *raddr, UDP_AUTH auth)
{
    char password[256];
	BYTE secret[256];
	// Usually 256, but not required.
	int secret_len = 256;
    
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        return 1;
    }
    
	if(auth == UDP_AUTH_NULL)
	{
		fprintf(stdout, "Press any key: ");
		fread(password, 1, 1, stdin);
		fprintf(stdout, "\n");
	}
	else
	{
		// Secure password retreival algorithm from enc
		get_password(password, 256, 1);
	}
	
	send_poke(s, raddr);
	int ret = dh_keyex(secret, &secret_len, s, raddr);
	if(ret != 0)
	{
		return ret;
	}

	ret = authenticate(secret, secret_len, password, s, raddr, auth);
	if(ret != 0)
	{
		return ret;
	}
    // Receive file.
    while(1)
    {
        char enc_data[1024 + 52];
        int recv_size = -1;
        UDP_CONTROL c = recv_data(enc_data, &recv_size, s, (struct sockaddr *)raddr);
        fprintf(stderr, "Received(%i)\n", recv_size);
        if(recv_size <= 0 || c != UDP_RECVFROM)
        {
            fprintf(stderr, "Received %s (%i)\n", udp_control_str(c), recv_size);
            break;
        }
        if(c == UDP_RECVFROM)
        {
            int data_size = -1;
            char dec_data[1024];
            // TODO: Check for Salted__ before you decrypt?
            int ret = decrypt_packet(enc_data, recv_size, secret, secret_len, dec_data, &data_size);
            // If they send /quit encrypted, then quit.
            // FIXME: What if the last packet is /quit? It's possible, unlikely.
            if(data_size == 5)
            {
                if(strncmp(dec_data, "/quit", 5) == 0)
                {
                    if(VERBOSE) fprintf(stderr, "Got a specific remote quit command.\n");
                    break;
                }
            }
            // Write the data to the specified file.
            if(ret == 0)
            {
                fwrite(dec_data, data_size, 1, data_io);
            }
            else
            {
                fprintf(stderr, "Error with received data: %i\n", ret);
            }
        }
    }
    return 0;
}

void err1(int value, const char *data)
{
    printf("Error: %s;\n", data);
    exit(value);
}

int usage()
{
    printf("Usage:   udp_sess1 [-c] [-r] [-f file] [-s] local_port remote_addr remote_port\n");
    printf("         -c      : Chat only, no files.\n");
    printf("         -r      : Receive file.\n");
    printf("         -f file : Read or Save file in filesystem.\n");
    printf("         The default options are to send a file from stdin.\n");
    printf("         All data is encrypt with a password using blowfish-cbc.\n");
    printf("Example: udp_sess1 3111 207.244.153.137 3222 < r.txt\n");
    printf("         udp_sess1 -f d.txt 3111 207.244.153.137 3222\n");
    printf("         udp_sess1 -r 3111 207.244.153.137 3222 > d.txt\n");
    printf("         udp_sess1 -c 3111 207.244.153.137 3222\n");
    printf("         udp_sess1 -c -s 3111 207.244.153.137 3222\n");
    return 1;
}

int main(int argc, char **argv)
{
    struct sockaddr_in raddr;
    struct sockaddr_in saddr;
    int s;
    int ret;
    
    int chat = 0;
    int send = 1;
    char *filename = 0;
    int dargs = 0;
    int local_io_port = 0;
    char local_io_addr[] = "0.0.0.0";
    int remote_recv_port = 0;
    char *remote_recv_addr = 0;
	UDP_AUTH auth = UDP_AUTH_PASSWORD;

    if(argc < 4) return usage();
    
    int i;
    for(i = 1; i < argc; i++)
    {
        if(strcmp(argv[i], "-c") == 0)
        {
            chat = 1;
        }
        else if(strcmp(argv[i], "-r") == 0)
        {
            send = 0;
        }
        else if(strcmp(argv[i], "-f") == 0)
        {
            // -f file
            i++;
            filename = argv[i];
        }
        else if(strcmp(argv[i], "-n") == 0)
        {
            // -n Null authentication
			auth = UDP_AUTH_NULL;
        }
        else if(strcmp(argv[i], "-d") == 0)
        {
            // -d DSA authentication
			auth = UDP_AUTH_DSA;
        }
        else
        {
            if(dargs == 0)
            {
                //fprintf(stderr, "setting local io port: %s, %i\n", argv[i], atoi(argv[i]));
                local_io_port = atoi(argv[i]);
            }
            if(dargs == 1)
            {
                //fprintf(stderr, "setting remote recv addr: %s\n", argv[i]);
                remote_recv_addr = argv[i];
            }
            if(dargs == 2)
            {
                //fprintf(stderr, "setting remote recv port: %s, %i\n", argv[i], atoi(argv[i]));
                remote_recv_port = atoi(argv[i]);
            }
            dargs++;
        }
    }
    // It is an error condition to not give all three data args.
    if(dargs < 3)
    {
        return usage();
    }
    
    FILE *data_io;
    if(filename == 0)
    {
        if(send == 1)
        {
            data_io = stdin;
        }
        else
        {
            data_io = stdout;
        }
    }
    else
    {
        if(send == 1)
        {
            data_io = fopen(filename, "rb");
        }
        else
        {
            data_io = fopen(filename, "wb");
        }
    }
    
    s = socket(PF_INET, SOCK_DGRAM, 0);
    if(s < 0) err1(1, "socket: SOCK_DGRAM");

    memset(&saddr, 0, sizeof(struct sockaddr_in));
    saddr.sin_family = AF_INET;
    saddr.sin_port = htons(local_io_port);
    saddr.sin_addr.s_addr = inet_addr(local_io_addr);

    /*
	struct hostent     *hp;
	if ((hp = gethostbyname(hostname)) == NULL) { // do we know the host's //
    errno = ECONNREFUSED;                       // address? //
    return(-1);                                // no //
    }
    
    memset(&sa,0,sizeof(sa));
    memcpy((char *)&sa.sin_addr, hp->h_addr, hp->h_length);     // set address //
    sa.sin_family = hp->h_addrtype;*/
    
	memset(&raddr, 0, sizeof(struct sockaddr));
    raddr.sin_family = AF_INET;
    raddr.sin_port = htons(remote_recv_port);
    raddr.sin_addr.s_addr = inet_addr(remote_recv_addr);
    
    //fprintf(stderr, "bind(%s, %i)\n", local_io_addr, local_io_port);
    ret = bind(s, (struct sockaddr *)&saddr, sizeof(struct sockaddr_in));
    if(ret) err1(1, "bind");

    //fprintf(stderr, "will send to(%s, %i) %08x\n", remote_recv_addr, remote_recv_port, &raddr);
    
    if(chat == 1)
    {
        ret = chat_main(s, &saddr, &raddr, auth);
    }
    else if(send == 1)
    {
        ret = send_main(data_io, s, &saddr, &raddr, auth);
    }
    else //if(send == 0)
    {
        ret = recv_main(data_io, s, &saddr, &raddr, auth);
    }
    
    fclose(data_io);
    shutdown(s, SHUT_WR);
    return ret;
}
